/* cnic.c: Broadcom CNIC core network driver.
 *
 * Copyright (c) 2006-2007 Broadcom Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation.
 *
 * Written by: John(Zongxi) Chen (zongxi@broadcom.com)
 * Modified and maintained by: Michael Chan <mchan@broadcom.com>
 */

#include <linux/version.h>
#if (LINUX_VERSION_CODE < 0x020612)
#include <linux/config.h>
#endif

#include <linux/module.h>

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/list.h>
#include <linux/slab.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <linux/in.h>
#include <linux/dma-mapping.h>
#include <asm/byteorder.h>
#include <linux/delay.h>
#include <linux/ethtool.h>
#include <linux/if_vlan.h>
#define BCM_VLAN 1
#include <net/ip.h>
#include <net/tcp.h>
#include <net/arp.h>
#include <net/neighbour.h>
#include <net/route.h>
#ifdef HAVE_NETEVENT
#include <net/netevent.h>
#endif

#include "bnx2.h"
#include "cnic_if.h"
#include "cnic.h"
#include "cnic_cm.h"

#define DRV_MODULE_NAME 	"cnic"
#define PFX DRV_MODULE_NAME ": "

static char version[] __devinitdata =
	"Broadcom NetXtreme II CNIC Driver " DRV_MODULE_NAME " v" CNIC_MODULE_VERSION " (" CNIC_MODULE_RELDATE ")\n";

MODULE_AUTHOR("John(Zongxi) Chen <zongxic@broadcom.com>");
MODULE_DESCRIPTION("Broadcom NetXtreme II CNIC Driver");
MODULE_LICENSE("GPL");
MODULE_VERSION(CNIC_MODULE_VERSION);

static LIST_HEAD(cnic_dev_list);
static DEFINE_RWLOCK(cnic_dev_lock);
static DEFINE_MUTEX(cnic_lock);

struct cnic_ulp_ops *cnic_ulp_tbl[MAX_CNIC_ULP_TYPE];

static int cnic_service_kq(void *, struct status_block *);
static void cnic_service_stop(void *);
static void cnic_service_start(void *);

static struct cnic_ops my_cnic_ops = {
	.cnic_owner	= THIS_MODULE,
	.cnic_handler	= cnic_service_kq,
	.cnic_stop	= cnic_service_stop,
	.cnic_start	= cnic_service_start,
};

static inline void cnic_hold(struct cnic_dev *dev)
{
	atomic_inc(&dev->ref_count);
}

static inline void cnic_put(struct cnic_dev *dev)
{
	atomic_dec(&dev->ref_count);
}

int cnic_register_driver(int ulp_type, struct cnic_ulp_ops *ulp_ops)
{
	struct cnic_dev *dev;

	if (ulp_type >= MAX_CNIC_ULP_TYPE) {
		printk(KERN_ERR PFX "cnic_register_driver: Bad type %d\n",
		       ulp_type);
		return -EINVAL;
	}
	mutex_lock(&cnic_lock);
	if (cnic_ulp_tbl[ulp_type]) {
		printk(KERN_ERR PFX "cnic_register_driver: Type %d has already "
				    "been registered\n", ulp_type);
		mutex_unlock(&cnic_lock);
		return -EBUSY;
	}

	read_lock(&cnic_dev_lock);
	list_for_each_entry(dev, &cnic_dev_list, list) {
		struct cnic_local *cp = dev->cnic_priv;

		clear_bit(ULP_F_INIT, &cp->ulp_flags[ulp_type]);
	}
	read_unlock(&cnic_dev_lock);

	rcu_assign_pointer(cnic_ulp_tbl[ulp_type], ulp_ops);
	mutex_unlock(&cnic_lock);

	read_lock(&cnic_dev_lock);
	list_for_each_entry(dev, &cnic_dev_list, list) {
		struct cnic_local *cp = dev->cnic_priv;

		/* Prevent race conditions with netdev_event */
		rtnl_lock();
		if (!test_and_set_bit(ULP_F_INIT, &cp->ulp_flags[ulp_type]))
			ulp_ops->cnic_init(dev);
		rtnl_unlock();
	}
	read_unlock(&cnic_dev_lock);

	return 0;
}

int cnic_unregister_driver(int ulp_type)
{
	struct cnic_dev *dev;

	if (ulp_type >= MAX_CNIC_ULP_TYPE) {
		printk(KERN_ERR PFX "cnic_unregister_driver: Bad type %d\n",
		       ulp_type);
		return -EINVAL;
	}
	mutex_lock(&cnic_lock);
	if (!cnic_ulp_tbl[ulp_type]) {
		printk(KERN_ERR PFX "cnic_unregister_driver: Type %d has not "
				    "been registered\n", ulp_type);
		goto out_unlock;
	}
	read_lock(&cnic_dev_lock);
	list_for_each_entry(dev, &cnic_dev_list, list) {
		struct cnic_local *cp = dev->cnic_priv;
		
		if (rcu_dereference(cp->ulp_ops[ulp_type])) {
			printk(KERN_ERR PFX "cnic_unregister_driver: Type %d "
			       "still has devices registered\n", ulp_type);
			read_unlock(&cnic_dev_lock);
			goto out_unlock;
		}
	}
	read_unlock(&cnic_dev_lock);

	rcu_assign_pointer(cnic_ulp_tbl[ulp_type], NULL);

	mutex_unlock(&cnic_lock);
	synchronize_rcu();
	return 0;

out_unlock:
	mutex_unlock(&cnic_lock);
	return -EINVAL;
}

EXPORT_SYMBOL(cnic_register_driver);
EXPORT_SYMBOL(cnic_unregister_driver);

static int cnic_start_hw(struct cnic_dev *);
static void cnic_stop_hw(struct cnic_dev *);

static int cnic_register_device(struct cnic_dev *dev, int ulp_type,
				void *ulp_ctx)
{
	struct cnic_local *cp = dev->cnic_priv;
	struct cnic_ulp_ops *ulp_ops;

	if (ulp_type >= MAX_CNIC_ULP_TYPE) {
		printk(KERN_ERR PFX "cnic_register_device: Bad type %d\n",
		       ulp_type);
		return -EINVAL;
	}
	mutex_lock(&cnic_lock);
	if (cnic_ulp_tbl[ulp_type] == NULL) {
		printk(KERN_ERR PFX "cnic_register_device: Driver with type %d "
				    "has not been registered\n", ulp_type);
		mutex_unlock(&cnic_lock);
		return -EAGAIN;
	}
	if (rcu_dereference(cp->ulp_ops[ulp_type])) {
		printk(KERN_ERR PFX "cnic_register_device: Type %d has already "
		       "been registered to this device\n", ulp_type);
		mutex_unlock(&cnic_lock);
		return -EBUSY;
	}
	if (!try_module_get(cnic_ulp_tbl[ulp_type]->owner)) {
		mutex_unlock(&cnic_lock);
		return -EBUSY;
	}

	clear_bit(ULP_F_START, &cp->ulp_flags[ulp_type]);
	cp->ulp_handle[ulp_type] = ulp_ctx;
	ulp_ops = cnic_ulp_tbl[ulp_type];
	rcu_assign_pointer(cp->ulp_ops[ulp_type], ulp_ops);
	cnic_hold(dev);
	if (!dev->use_count) {
		if (!test_bit(CNIC_F_IF_GOING_DOWN, &dev->flags)) {
			if (dev->netdev->flags & IFF_UP)
				set_bit(CNIC_F_IF_UP, &dev->flags);
		}
	}
	dev->use_count++;

	if (dev->use_count == 1) {
		if (test_bit(CNIC_F_IF_UP, &dev->flags))
			cnic_start_hw(dev);
	}

	if (test_bit(CNIC_F_CNIC_UP, &dev->flags))
		if (!test_and_set_bit(ULP_F_START, &cp->ulp_flags[ulp_type]))
			ulp_ops->cnic_start(cp->ulp_handle[ulp_type]);

	mutex_unlock(&cnic_lock);

	return 0;

}

static int cnic_unregister_device(struct cnic_dev *dev, int ulp_type)
{
	struct cnic_local *cp = dev->cnic_priv;

	if (ulp_type >= MAX_CNIC_ULP_TYPE) {
		printk(KERN_ERR PFX "cnic_unregister_device: Bad type %d\n",
		       ulp_type);
		return -EINVAL;
	}
	mutex_lock(&cnic_lock);
	if (rcu_dereference(cp->ulp_ops[ulp_type])) {
		dev->use_count--;
		module_put(cp->ulp_ops[ulp_type]->owner);
		rcu_assign_pointer(cp->ulp_ops[ulp_type], NULL);
		if (dev->use_count == 0)
			cnic_stop_hw(dev);
		cnic_put(dev);
	} else {
		printk(KERN_ERR PFX "cnic_unregister_device: device not "
		       "registered to this ulp type %d\n", ulp_type);
		mutex_unlock(&cnic_lock);
		return -EINVAL;
	}
	mutex_unlock(&cnic_lock);

	synchronize_rcu();

	return 0;
}

#ifdef CNIC_SYSFS
static ssize_t show_pci_bar(struct class_device *cdev, char *buf)
{
	struct cnic_dev *dev = container_of(cdev, struct cnic_dev, class_dev);

	return sprintf(buf, "0x%.8x\n", (u32)pci_resource_start(dev->pcidev, 0));
}

static ssize_t show_kq_intr_coal(struct class_device *cdev, char *buf)
{
	struct cnic_dev *dev = container_of(cdev, struct cnic_dev, class_dev);

	return sprintf(buf, "0x%.8x\n", REG_RD(dev, BNX2_HC_COMP_PROD_TRIP));
}

static ssize_t set_kq_intr_coal(struct class_device *cdev, const char *buf,
			      size_t count)
{
	struct cnic_dev *dev = container_of(cdev, struct cnic_dev, class_dev);
	u32 val;

	if (sscanf(buf, " 0x%x ", &val) > 0)
		REG_WR(dev, BNX2_HC_COMP_PROD_TRIP, val);
	return count;
}

static ssize_t show_kq_com_ticks(struct class_device *cdev, char *buf)
{
	struct cnic_dev *dev = container_of(cdev, struct cnic_dev, class_dev);

	return sprintf(buf, "0x%.8x\n", REG_RD(dev, BNX2_HC_COM_TICKS));
}

static ssize_t set_kq_com_ticks(struct class_device *cdev, const char *buf,
			      size_t count)
{
	struct cnic_dev *dev = container_of(cdev, struct cnic_dev, class_dev);
	u32 val;

	if (sscanf(buf, " 0x%x ", &val) > 0)
		REG_WR(dev, BNX2_HC_COM_TICKS, val);
	return count;
}

static ssize_t show_kq_cmd_ticks(struct class_device *cdev, char *buf)
{
	struct cnic_dev *dev = container_of(cdev, struct cnic_dev, class_dev);

	return sprintf(buf, "0x%.8x\n", REG_RD(dev, BNX2_HC_CMD_TICKS));
}

static ssize_t set_kq_cmd_ticks(struct class_device *cdev, const char *buf,
			      size_t count)
{
	struct cnic_dev *dev = container_of(cdev, struct cnic_dev, class_dev);
	u32 val;

	if (sscanf(buf, " 0x%x ", &val) > 0)
		REG_WR(dev, BNX2_HC_CMD_TICKS, val);
	return count;
}

static CLASS_DEVICE_ATTR(pci_bar, S_IRUGO, show_pci_bar, NULL);
static CLASS_DEVICE_ATTR(kq_intr_coal, S_IRUGO | S_IWUSR, show_kq_intr_coal,
			 set_kq_intr_coal);
static CLASS_DEVICE_ATTR(kq_com_ticks, S_IRUGO | S_IWUSR, show_kq_com_ticks,
			 set_kq_com_ticks);
static CLASS_DEVICE_ATTR(kq_cmd_ticks, S_IRUGO | S_IWUSR, show_kq_cmd_ticks,
			 set_kq_cmd_ticks);

static struct class_device_attribute *cnic_class_attributes[] = {
	&class_device_attr_pci_bar,
	&class_device_attr_kq_intr_coal,
	&class_device_attr_kq_com_ticks,
	&class_device_attr_kq_cmd_ticks
};

static void cnic_sysfs_release(struct class_device *class_dev)
{
}

static struct class cnic_class = {
	.name    = "cnic",
	.release = cnic_sysfs_release,
};

static int cnic_register_sysfs(struct cnic_dev *device)
{
	struct class_device *class_dev = &device->class_dev;
	char dev_name[BUS_ID_SIZE];
	int ret;
	int i;

	class_dev->class      = &cnic_class;
	class_dev->class_data = device;
	snprintf(dev_name, BUS_ID_SIZE, "%.2x:%.2x.%.1x",
		 device->pcidev->bus->number, PCI_SLOT(device->pcidev->devfn),
		 PCI_FUNC(device->pcidev->devfn));
	strlcpy(class_dev->class_id, dev_name, BUS_ID_SIZE);

	ret = class_device_register(class_dev);
	if (ret)
		goto err;

	for (i = 0; i < ARRAY_SIZE(cnic_class_attributes); ++i) {
		ret = class_device_create_file(class_dev,
					       cnic_class_attributes[i]);
		if (ret)
			goto err_unregister;
	}

	return 0;

err_unregister:
	class_device_unregister(class_dev);
err:
	return ret;
}

static void cnic_unregister_sysfs(struct cnic_dev *device)
{
	class_device_unregister(&device->class_dev);
}

static int cnic_sysfs_setup(void)
{
	return class_register(&cnic_class);
}

static void cnic_sysfs_cleanup(void)
{
	class_unregister(&cnic_class);
}
#else
static void cnic_unregister_sysfs(struct cnic_dev *device)
{
}

static int cnic_register_sysfs(struct cnic_dev *device)
{
	return 0;
}

static int cnic_sysfs_setup(void)
{
	return 0;
}

static void cnic_sysfs_cleanup(void)
{
}

#endif

static void cnic_free_resc(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;
	int i;

	for (i = 0; i < cp->ctx_pages; i++) {
		if (cp->ctx_arr[i].ctx) {
			pci_free_consistent(dev->pcidev, BCM_PAGE_SIZE,
					    cp->ctx_arr[i].ctx,
					    cp->ctx_arr[i].mapping);
			cp->ctx_arr[i].ctx = NULL;
		}
	}
	kfree(cp->ctx_arr);
	cp->ctx_arr = NULL;
	cp->ctx_pages = 0;

	if (cp->kwq_pgtbl) {
		pci_free_consistent(dev->pcidev, cp->kwq_pgtbl_size,
				    cp->kwq_pgtbl, cp->kwq_pgtbl_mapping);
		cp->kwq_pgtbl = NULL;
	}
	for (i = 0; i < KWQ_PAGE_CNT; i++) {
		if (cp->kwq[i]) {
			pci_free_consistent(dev->pcidev, BCM_PAGE_SIZE,
					    cp->kwq[i], cp->kwq_mapping[i]);
			cp->kwq[i] = NULL;
		}
	}
	if (cp->kcq_pgtbl) {
		pci_free_consistent(dev->pcidev, cp->kcq_pgtbl_size,
				    cp->kcq_pgtbl, cp->kcq_pgtbl_mapping);
		cp->kcq_pgtbl = NULL;
	}
	for (i = 0; i < KCQ_PAGE_CNT; i++) {
		if (cp->kcq[i]) {
			pci_free_consistent(dev->pcidev, BCM_PAGE_SIZE,
					    cp->kcq[i], cp->kcq_mapping[i]);
			cp->kcq[i] = NULL;
		}
	}

	return;
}

static int cnic_alloc_context(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;

	if (CHIP_NUM(cp) == CHIP_NUM_5709) {
		int i = 0, arr_size;
		u32 j;

		cp->cids_pp = PAGE_SIZE / 128;
		arr_size = BNX2_MAX_CID / cp->cids_pp * sizeof(struct cnic_ctx);
		cp->ctx_arr = kmalloc(arr_size, GFP_KERNEL);
		if (cp->ctx_arr == NULL)
			return -ENOMEM;

		memset(cp->ctx_arr, 0, arr_size);

		for (j = 0x2100; j <= 0x2980; j += cp->cids_pp, i++)
			cp->ctx_arr[i].cid = j;
		for (j = 0x3300; j <= 0x3380; j += cp->cids_pp, i++)
			cp->ctx_arr[i].cid = j;

		cp->ctx_pages = i;

		for (i = 0; i < cp->ctx_pages; i++) {
			cp->ctx_arr[i].ctx =
				pci_alloc_consistent(dev->pcidev, BCM_PAGE_SIZE,
						     &cp->ctx_arr[i].mapping);
			if (cp->ctx_arr[i].ctx == NULL)
				return -ENOMEM;
		}
	}
	return 0;
}

static int cnic_alloc_resc(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;
	int i;

	cp->kwq_pgtbl_size = ((KWQ_PAGE_CNT * 8) + BCM_PAGE_SIZE - 1) &
			      (BCM_PAGE_SIZE - 1);
	cp->kwq_pgtbl = pci_alloc_consistent(dev->pcidev, cp->kwq_pgtbl_size,
					     &cp->kwq_pgtbl_mapping);
	if (cp->kwq_pgtbl == NULL)
		goto error;

	for (i = 0; i < KWQ_PAGE_CNT; i++) {
		cp->kwq[i] = pci_alloc_consistent(dev->pcidev, BCM_PAGE_SIZE,
						  &cp->kwq_mapping[i]);
		if (cp->kwq[i] == NULL)
			goto error;
	}

	cp->kcq_pgtbl_size = ((KCQ_PAGE_CNT * 8) + BCM_PAGE_SIZE - 1) &
			     (BCM_PAGE_SIZE - 1);
	cp->kcq_pgtbl = pci_alloc_consistent(dev->pcidev, cp->kcq_pgtbl_size,
					     &cp->kcq_pgtbl_mapping);
	if (cp->kcq_pgtbl == NULL)
		goto error;

	for (i = 0; i < KCQ_PAGE_CNT; i++) {
		cp->kcq[i] = pci_alloc_consistent(dev->pcidev, BCM_PAGE_SIZE,
						  &cp->kcq_mapping[i]);
		if (cp->kcq[i] == NULL)
			goto error;
	}

	if (cnic_alloc_context(dev))
		goto error;

	return 0;
	
error:
	cnic_free_resc(dev);
	return -ENOMEM;
}

static void cnic_setup_page_tbl(u32 *page_table, u32 page_cnt,
				dma_addr_t base_mapping[])
{
	int i;

	for (i = 0; i < page_cnt; i++) {
		/* Each entry needs to be in big endian format. */
		*page_table = (u32) ((u64) base_mapping[i] >> 32);
		page_table++;
		*page_table = (u32) base_mapping[i];
		page_table++;
	}
}

static inline u32 cnic_kwq_avail(struct cnic_local *cp)
{
	return (MAX_KWQ_IDX -
		((cp->kwq_prod_idx - cp->kwq_con_idx) & MAX_KWQ_IDX));
}

static int cnic_submit_kernel_wqes(struct cnic_dev *dev, struct kwqe *wqes[],
			    u32 num_wqes)
{
	struct cnic_local *cp = dev->cnic_priv;
	struct kwqe *prod_qe;
	u16 prod, sw_prod, i;

	if (!test_bit(CNIC_F_CNIC_UP, &dev->flags))
		return -EAGAIN;		/* bnx2 is down */

	spin_lock_bh(&cp->cnic_ulp_lock);
	if (num_wqes > cnic_kwq_avail(cp)) {
		spin_unlock_bh(&cp->cnic_ulp_lock);
		return -EAGAIN;
	}

	prod = cp->kwq_prod_idx;
	sw_prod = prod & MAX_KWQ_IDX;
	for (i = 0; i < num_wqes; i++) {
		prod_qe = &cp->kwq[KWQ_PG(sw_prod)][KWQ_IDX(sw_prod)];
		memcpy(prod_qe, wqes[i], sizeof(struct kwqe));
		prod++;
		sw_prod = prod & MAX_KWQ_IDX;
	}
	cp->kwq_prod_idx = prod;

	REG_WR16(dev, cp->kwq_io_addr, cp->kwq_prod_idx);

	spin_unlock_bh(&cp->cnic_ulp_lock);
	return 0;
}

static void service_kcqes(struct cnic_dev *dev, int num_cqes)
{
	struct cnic_local *cp = dev->cnic_priv;
	int i, j;

	i = 0;
	j = 1;
	while (num_cqes) {
		struct cnic_ulp_ops *ulp_ops;
		int ulp_type;
		u32 kcqe_op_flag = cp->completed_kcq[i]->kcqe_op_flag &
				   KCQE_FLAGS_LAYER_MASK;

		while (j < num_cqes) {
			if ((cp->completed_kcq[i + j]->kcqe_op_flag &
			     KCQE_FLAGS_LAYER_MASK) != kcqe_op_flag) {
				break;
			}
			j++;
		}

		if (kcqe_op_flag == KCQE_FLAGS_LAYER_MASK_L5_RDMA)
			ulp_type = CNIC_ULP_RDMA;
		else if (kcqe_op_flag == KCQE_FLAGS_LAYER_MASK_L5_ISCSI)
			ulp_type = CNIC_ULP_ISCSI;
		else if (kcqe_op_flag == KCQE_FLAGS_LAYER_MASK_L4)
			ulp_type = CNIC_ULP_L4;
		else {
			printk(KERN_ERR PFX "%s: Unknown type of KCQE(0x%x)\n",
			       dev->netdev->name, kcqe_op_flag);
			goto end;
		}

		rcu_read_lock();
		ulp_ops = rcu_dereference(cp->ulp_ops[ulp_type]);
		if (likely(ulp_ops)) {
			ulp_ops->indicate_kcqes(cp->ulp_handle[ulp_type],
						  cp->completed_kcq + i, j);
		}
		rcu_read_unlock();
end:
		num_cqes -= j;
		i += j;
		j = 1;
	}
	return;
}

static int cnic_get_kcqes(struct cnic_dev *dev, u16 hw_prod, u16 sw_prod)
{
	struct cnic_local *cp = dev->cnic_priv;
	u16 i, max;
	struct kcqe *kcqe;
	int kcqe_cnt = 0;

	i = sw_prod & MAX_KCQ_IDX;
	max = hw_prod & MAX_KCQ_IDX;
	while ((i != max) && (kcqe_cnt < MAX_COMPLETED_KCQE)) {
		cp->completed_kcq[kcqe_cnt++] = &cp->kcq[KCQ_PG(i)][KCQ_IDX(i)];
		i = (i + 1) & MAX_KCQ_IDX;
	}

	kcqe = cp->completed_kcq[kcqe_cnt - 1];
	while (kcqe->kcqe_op_flag & KCQE_FLAGS_NEXT) {
		kcqe_cnt--;
		if (kcqe_cnt == 0)
			goto done;
		kcqe = cp->completed_kcq[kcqe_cnt - 1];
	}
done:
	return kcqe_cnt;
}

static int cnic_service_kq(void *data, struct status_block *status_blk)
{
	struct cnic_dev *dev = data;
	struct cnic_local *cp = dev->cnic_priv;
	u32 status_idx = status_blk->status_idx;
	u16 hw_prod, sw_prod;
	int kcqe_cnt;

	cp->kwq_con_idx = status_blk->status_rx_quick_consumer_index15;

	hw_prod = status_blk->status_completion_producer_index;
	sw_prod = cp->kcq_prod_idx;
	while (sw_prod != hw_prod) {
		kcqe_cnt = cnic_get_kcqes(dev, hw_prod, sw_prod);
		if (kcqe_cnt == 0)
			goto done;
		sw_prod += kcqe_cnt;

		service_kcqes(dev, kcqe_cnt);

		/* Tell compiler that status_blk fields can change. */
		barrier();
		if (status_idx != status_blk->status_idx) {
			status_idx = status_blk->status_idx;
			cp->kwq_con_idx =
				status_blk->status_rx_quick_consumer_index15;
			hw_prod = status_blk->status_completion_producer_index;
		} else
			break;
	}

done:
	REG_WR16(dev, cp->kcq_io_addr, sw_prod);

	cp->kcq_prod_idx = sw_prod;
	return status_idx;
}

#ifdef CONFIG_PCI_MSI
static void cnic_service_kq_msix(unsigned long data)
{
	struct cnic_dev *dev = (struct cnic_dev *) data;
	struct cnic_local *cp = dev->cnic_priv;
	struct status_block_msix *status_blk = cp->status_blk;
	u32 status_idx = status_blk->status_idx;
	u16 hw_prod, sw_prod;
	int kcqe_cnt;

	cp->kwq_con_idx = status_blk->status_cmd_consumer_index;

	hw_prod = status_blk->status_completion_producer_index;
	sw_prod = cp->kcq_prod_idx;
	while (sw_prod != hw_prod) {
		kcqe_cnt = cnic_get_kcqes(dev, hw_prod, sw_prod);
		if (kcqe_cnt == 0)
			goto done;
		sw_prod += kcqe_cnt;

		service_kcqes(dev, kcqe_cnt);

		/* Tell compiler that status_blk fields can change. */
		barrier();
		if (status_idx != status_blk->status_idx) {
			status_idx = status_blk->status_idx;
			cp->kwq_con_idx = status_blk->status_cmd_consumer_index;
			hw_prod = status_blk->status_completion_producer_index;
		} else
			break;
	}

done:
	REG_WR16(dev, cp->kcq_io_addr, sw_prod);

	cp->kcq_prod_idx = sw_prod;
	cp->last_status_idx = status_idx;
	REG_WR(dev, BNX2_PCICFG_INT_ACK_CMD, (BNX2_CNIC_VEC << 24) |
	       BNX2_PCICFG_INT_ACK_CMD_INDEX_VALID | cp->last_status_idx);
}

#if (LINUX_VERSION_CODE >= 0x20613)
static irqreturn_t cnic_msix(int irq, void *dev_instance)
#else
static irqreturn_t cnic_msix(int irq, void *dev_instance, struct pt_regs *regs)
#endif
{
	struct cnic_dev *dev = dev_instance;
	struct cnic_local *cp = dev->cnic_priv;
	u16 prod = cp->kcq_prod_idx & MAX_KCQ_IDX;

	prefetch(cp->status_blk);
	prefetch(&cp->kcq[KCQ_PG(prod)][KCQ_IDX(prod)]);

	if (likely(test_bit(CNIC_F_CNIC_UP, &dev->flags)))
		tasklet_schedule(&cp->cnic_msix_task);

	return IRQ_HANDLED;
}
#endif

static void cnic_ulp_stop(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;
	int if_type;

	rcu_read_lock();
	for (if_type = 0; if_type < MAX_CNIC_ULP_TYPE; if_type++) {
		struct cnic_ulp_ops *ulp_ops;

		ulp_ops = rcu_dereference(cp->ulp_ops[if_type]);
		if (!ulp_ops)
			continue;

		if (test_and_clear_bit(ULP_F_START, &cp->ulp_flags[if_type]))
			ulp_ops->cnic_stop(cp->ulp_handle[if_type]);
	}
	rcu_read_unlock();
}

static void cnic_ulp_start(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;
	int if_type;

	rcu_read_lock();
	for (if_type = 0; if_type < MAX_CNIC_ULP_TYPE; if_type++) {
		struct cnic_ulp_ops *ulp_ops;

		ulp_ops = rcu_dereference(cp->ulp_ops[if_type]);
		if (!ulp_ops)
			continue;

		if (!test_and_set_bit(ULP_F_START, &cp->ulp_flags[if_type]))
			ulp_ops->cnic_start(cp->ulp_handle[if_type]);
	}
	rcu_read_unlock();
}

static void cnic_service_start(void *data)
{
	struct cnic_dev *dev = data;

	cnic_hold(dev);
	mutex_lock(&cnic_lock);
	set_bit(CNIC_F_IF_UP, &dev->flags);
	if (dev->use_count) {
		if (!cnic_start_hw(dev))
			cnic_ulp_start(dev);
	}
	mutex_unlock(&cnic_lock);
	cnic_put(dev);
}

static void cnic_service_stop(void *data)
{
	struct cnic_dev *dev = data;

	cnic_hold(dev);
	mutex_lock(&cnic_lock);
	clear_bit(CNIC_F_IF_UP, &dev->flags);
	cnic_ulp_stop(dev);
	cnic_stop_hw(dev);
	mutex_unlock(&cnic_lock);
	cnic_put(dev);
}

static void cnic_ulp_init(struct cnic_dev *dev)
{
	int i;
	struct cnic_local *cp = dev->cnic_priv;

	rcu_read_lock();
	for (i = 0; i < MAX_CNIC_ULP_TYPE_EXT; i++) {
		struct cnic_ulp_ops *ulp_ops;

		ulp_ops = rcu_dereference(cnic_ulp_tbl[i]);
		if (!ulp_ops || !try_module_get(ulp_ops->owner))
			continue;

		if (!test_and_set_bit(ULP_F_INIT, &cp->ulp_flags[i]))
			ulp_ops->cnic_init(dev);

		module_put(ulp_ops->owner);
	}
	rcu_read_unlock();
}

static void cnic_ulp_exit(struct cnic_dev *dev)
{
	int i;
	struct cnic_local *cp = dev->cnic_priv;

	rcu_read_lock();
	for (i = 0; i < MAX_CNIC_ULP_TYPE_EXT; i++) {
		struct cnic_ulp_ops *ulp_ops;

		ulp_ops = rcu_dereference(cnic_ulp_tbl[i]);
		if (!ulp_ops || !try_module_get(ulp_ops->owner))
			continue;

		if (test_and_clear_bit(ULP_F_INIT, &cp->ulp_flags[i]))
			ulp_ops->cnic_exit(dev);

		module_put(ulp_ops->owner);
	}
	rcu_read_unlock();
}

static int cnic_queue_work(struct cnic_local *cp, u32 work_type, void *data)
{
	struct cnic_work_node *node;
	int bytes = sizeof(u32 *);

	spin_lock_bh(&cp->cm_lock);

	node = &cp->cnic_work_ring[cp->cnic_wr_prod];
	node->work_type = work_type;
	if (work_type == WORK_TYPE_KCQE)
		bytes = sizeof(struct kcqe);
	if (work_type == WORK_TYPE_REDIRECT)
		bytes = sizeof(struct cnic_redirect_entry);
	memcpy(&node->work_data, data, bytes);
	cp->cnic_wr_prod++;
	cp->cnic_wr_prod &= WORK_RING_SIZE_MASK;

	spin_unlock_bh(&cp->cm_lock);
	return 0;
}

static int cnic_cm_offload_pg(struct cnic_sock *csk)
{
	struct cnic_dev *dev = csk->dev;
	struct l4_kwq_offload_pg *l4kwqe;
	struct kwqe *wqes[1];
	struct neighbour *neigh = csk->dst->neighbour;
	struct net_device *netdev = neigh->dev;

#ifndef HAVE_NETEVENT
	memcpy(csk->old_ha, &neigh->ha[0], 6);
#endif
	l4kwqe = (struct l4_kwq_offload_pg *) &csk->kwqe1;
	memset(l4kwqe, 0, sizeof(*l4kwqe));
	wqes[0] = (struct kwqe *) l4kwqe;

	l4kwqe->op_code = L4_KWQE_OPCODE_VALUE_OFFLOAD_PG;
	l4kwqe->flags =
		L4_LAYER_CODE << L4_KWQ_OFFLOAD_PG_LAYER_CODE_SHIFT;
	l4kwqe->l2hdr_nbytes = ETH_HLEN;
	l4kwqe->da0 = neigh->ha[0];
	l4kwqe->da1 = neigh->ha[1];
	l4kwqe->da2 = neigh->ha[2];
	l4kwqe->da3 = neigh->ha[3];
	l4kwqe->da4 = neigh->ha[4];
	l4kwqe->da5 = neigh->ha[5];

	l4kwqe->sa0 = netdev->dev_addr[0];
	l4kwqe->sa1 = netdev->dev_addr[1];
	l4kwqe->sa2 = netdev->dev_addr[2];
	l4kwqe->sa3 = netdev->dev_addr[3];
	l4kwqe->sa4 = netdev->dev_addr[4];
	l4kwqe->sa5 = netdev->dev_addr[5];

	l4kwqe->etype = ETH_P_IP;
	l4kwqe->ipid_count = DEF_IPID_COUNT;
	l4kwqe->host_opaque = csk->l5_cid;

	if (csk->vlan_id) {
		l4kwqe->pg_flags |= L4_KWQ_OFFLOAD_PG_VLAN_TAGGING;
		l4kwqe->vlan_tag = csk->vlan_id;
		l4kwqe->l2hdr_nbytes += 4;
	}

	return (dev->submit_kwqes(dev, wqes, 1));
}

static int cnic_cm_update_pg(struct cnic_sock *csk)
{
	struct cnic_dev *dev = csk->dev;
	struct l4_kwq_update_pg *l4kwqe;
	struct kwqe *wqes[1];
	struct neighbour *neigh = csk->dst->neighbour;

#ifndef HAVE_NETEVENT
	memcpy(csk->old_ha, &neigh->ha[0], 6);
#endif
	l4kwqe = (struct l4_kwq_update_pg *) &csk->kwqe1;
	memset(l4kwqe, 0, sizeof(*l4kwqe));
	wqes[0] = (struct kwqe *) l4kwqe;

	l4kwqe->opcode = L4_KWQE_OPCODE_VALUE_UPDATE_PG;
	l4kwqe->flags =
		L4_LAYER_CODE << L4_KWQ_UPDATE_PG_LAYER_CODE_SHIFT;
	l4kwqe->pg_cid = csk->pg_cid;
	l4kwqe->da0 = neigh->ha[0];
	l4kwqe->da1 = neigh->ha[1];
	l4kwqe->da2 = neigh->ha[2];
	l4kwqe->da3 = neigh->ha[3];
	l4kwqe->da4 = neigh->ha[4];
	l4kwqe->da5 = neigh->ha[5];

	l4kwqe->pg_host_opaque = 0;
	l4kwqe->pg_valids = L4_KWQ_UPDATE_PG_VALIDS_DA;

	return (dev->submit_kwqes(dev, wqes, 1));
}

static int cnic_cm_upload_pg(struct cnic_sock *csk)
{
	struct cnic_dev *dev = csk->dev;
	struct l4_kwq_upload *l4kwqe;
	struct kwqe *wqes[1];

	l4kwqe = (struct l4_kwq_upload *) &csk->kwqe1;
	memset(l4kwqe, 0, sizeof(*l4kwqe));
	wqes[0] = (struct kwqe *) l4kwqe;

	l4kwqe->opcode = L4_KWQE_OPCODE_VALUE_UPLOAD_PG;
	l4kwqe->flags =
		L4_LAYER_CODE << L4_KWQ_UPLOAD_LAYER_CODE_SHIFT;
	l4kwqe->cid = csk->pg_cid;

	return (dev->submit_kwqes(dev, wqes, 1));
}

#ifdef HAVE_NETEVENT
static void cnic_redirect(struct cnic_local *cp, struct dst_entry *new,
			  struct dst_entry *old)
{
	int i, found = 0;

	spin_lock_bh(&cp->cm_lock);
	for (i = 0; i < MAX_CM_SK_TBL_SZ; i++) {
		struct cnic_sock *csk;

		csk = &cp->csk_tbl[i];
		if (test_bit(SK_F_INUSE, &csk->flags) && csk->dst == old)
			found = 1;
	}
	spin_unlock_bh(&cp->cm_lock);

	if (found) {
		struct cnic_redirect_entry cnic_redir;

		dst_hold(new);
		cnic_redir.old_dst = old;
		cnic_redir.new_dst = new;
		cnic_queue_work(cp, WORK_TYPE_REDIRECT, &cnic_redir);
		tasklet_schedule(&cp->cnic_task);
	}
}

static void cnic_update_neigh(struct cnic_local *cp, struct neighbour *neigh)
{
	int i, found = 0;

	spin_lock_bh(&cp->cm_lock);
	for (i = 0; i < MAX_CM_SK_TBL_SZ; i++) {
		struct cnic_sock *csk;

		csk = &cp->csk_tbl[i];
		if (test_bit(SK_F_INUSE, &csk->flags) && csk->dst) {
			if (csk->dst->neighbour == neigh) {
				found = 1;
				break;
			}
		}
	}
	spin_unlock_bh(&cp->cm_lock);

	if (!found)
		return;

	neigh_hold(neigh);

	cnic_queue_work(cp, WORK_TYPE_NEIGH_UPDATE, &neigh);
	tasklet_schedule(&cp->cnic_task);
}

static int cnic_net_callback(struct notifier_block *this, unsigned long event,
	void *ptr)
{
	struct cnic_local *cp = container_of(this, struct cnic_local, cm_nb);

	if (event == NETEVENT_NEIGH_UPDATE) {
		struct neighbour *neigh = ptr;

		cnic_update_neigh(cp, neigh);

	} else if (event == NETEVENT_REDIRECT) {
		struct netevent_redirect *netevent = ptr;
		struct dst_entry *old_dst = netevent->old;
		struct dst_entry *new_dst = netevent->new;

		cnic_redirect(cp, new_dst, old_dst);
	}
	return 0;
}
#endif

static int cnic_ok_to_connect(struct cnic_sock *csk)
{
	if (test_bit(SK_F_INUSE, &csk->flags) &&
	    !test_bit(SK_F_OFFLD_PENDING, &csk->flags) &&
	    !test_bit(SK_F_OFFLD_COMPLETE, &csk->flags) &&
	    test_bit(SK_F_PG_OFFLD_COMPLETE, &csk->flags))
		return 1;
	return 0;
}

static int cnic_cm_conn_req(struct cnic_sock *csk)
{
	struct cnic_dev *dev = csk->dev;
	struct l4_kwq_connect_req1 *l4kwqe1;
	struct l4_kwq_connect_req3 *l4kwqe3;
	struct kwqe *wqes[2];
	u8 tcp_flags = 0;

	l4kwqe1 = (struct l4_kwq_connect_req1 *) &csk->kwqe2;
	l4kwqe3 = (struct l4_kwq_connect_req3 *) &csk->kwqe3;
	memset(l4kwqe1, 0, sizeof(*l4kwqe1));
	memset(l4kwqe3, 0, sizeof(*l4kwqe3));
	wqes[0] = (struct kwqe *) l4kwqe1;
	wqes[1] = (struct kwqe *) l4kwqe3;

	l4kwqe1->op_code = L4_KWQE_OPCODE_VALUE_CONNECT1;
	l4kwqe1->flags =
		(L4_LAYER_CODE << L4_KWQ_CONNECT_REQ1_LAYER_CODE_SHIFT) |
		 L4_KWQ_CONNECT_REQ3_LINKED_WITH_NEXT;
	l4kwqe1->cid = csk->cid;
	l4kwqe1->pg_cid = csk->pg_cid;
	l4kwqe1->src_ip = be32_to_cpu(csk->src_ip);
	l4kwqe1->dst_ip = be32_to_cpu(csk->dst_ip);
	l4kwqe1->src_port = be16_to_cpu(csk->src_port);
	l4kwqe1->dst_port = be16_to_cpu(csk->dst_port);
	if (test_bit(SK_TCP_NO_DELAY_ACK, &csk->tcp_flags))
		tcp_flags |= L4_KWQ_CONNECT_REQ1_NO_DELAY_ACK;
	if (test_bit(SK_TCP_KEEP_ALIVE, &csk->tcp_flags))
		tcp_flags |= L4_KWQ_CONNECT_REQ1_KEEP_ALIVE;
	if (test_bit(SK_TCP_NAGLE, &csk->tcp_flags))
		tcp_flags |= L4_KWQ_CONNECT_REQ1_NAGLE_ENABLE;
	if (test_bit(SK_TCP_TIMESTAMP, &csk->tcp_flags))
		tcp_flags |= L4_KWQ_CONNECT_REQ1_TIME_STAMP;
	if (test_bit(SK_TCP_SACK, &csk->tcp_flags))
		tcp_flags |= L4_KWQ_CONNECT_REQ1_SACK;
	if (test_bit(SK_TCP_SEG_SCALING, &csk->tcp_flags))
		tcp_flags |= L4_KWQ_CONNECT_REQ1_SEG_SCALING;
	
	l4kwqe1->tcp_flags = tcp_flags;

	l4kwqe3->op_code = L4_KWQE_OPCODE_VALUE_CONNECT3;
	l4kwqe3->flags =
		L4_LAYER_CODE << L4_KWQ_CONNECT_REQ3_LAYER_CODE_SHIFT;
	l4kwqe3->ka_timeout = csk->ka_timeout;
	l4kwqe3->ka_interval = csk->ka_interval;
	l4kwqe3->ka_max_probe_count = csk->ka_max_probe_count;
	l4kwqe3->tos = csk->tos;
	l4kwqe3->ttl = csk->ttl;
	l4kwqe3->snd_seq_scale = csk->snd_seq_scale;
	l4kwqe3->pmtu = dst_mtu(csk->dst);
	l4kwqe3->mss = l4kwqe3->pmtu - 40;
	l4kwqe3->rcv_buf = csk->rcv_buf;
	l4kwqe3->snd_buf = csk->snd_buf;
	l4kwqe3->seed = csk->seed;

	return (dev->submit_kwqes(dev, wqes, 2));
}

static int cnic_cm_close_req(struct cnic_sock *csk)
{
	struct cnic_dev *dev = csk->dev;
	struct l4_kwq_close_req *l4kwqe;
	struct kwqe *wqes[1];

	l4kwqe = (struct l4_kwq_close_req *) &csk->kwqe2;
	memset(l4kwqe, 0, sizeof(*l4kwqe));
	wqes[0] = (struct kwqe *) l4kwqe;

	l4kwqe->op_code = L4_KWQE_OPCODE_VALUE_CLOSE;
	l4kwqe->flags = L4_LAYER_CODE << L4_KWQ_CLOSE_REQ_LAYER_CODE_SHIFT;
	l4kwqe->cid = csk->cid;

	return (dev->submit_kwqes(dev, wqes, 1));
}

static int cnic_cm_abort_req(struct cnic_sock *csk)
{
	struct cnic_dev *dev = csk->dev;
	struct l4_kwq_reset_req *l4kwqe;
	struct kwqe *wqes[1];

	l4kwqe = (struct l4_kwq_reset_req *) &csk->kwqe2;
	memset(l4kwqe, 0, sizeof(*l4kwqe));
	wqes[0] = (struct kwqe *) l4kwqe;

	l4kwqe->op_code = L4_KWQE_OPCODE_VALUE_RESET;
	l4kwqe->flags = L4_LAYER_CODE << L4_KWQ_RESET_REQ_LAYER_CODE_SHIFT;
	l4kwqe->cid = csk->cid;

	return (dev->submit_kwqes(dev, wqes, 1));
}

static int cnic_cm_create(struct cnic_dev *dev, int ulp_type, u32 cid,
			  u32 l5_cid, struct cnic_sock **csk, void *context)
{
	struct cnic_local *cp = dev->cnic_priv;
	struct cnic_sock *csk1;

	if (l5_cid >= MAX_CM_SK_TBL_SZ)
		return -EINVAL;
		
	csk1 = &cp->csk_tbl[l5_cid];
	if (test_and_set_bit(SK_F_INUSE, &csk1->flags))
		return -EINVAL;

	csk1->dev = dev;
	csk1->cid = cid;
	csk1->l5_cid = l5_cid;
	csk1->ulp_type = ulp_type;
	csk1->context = context;

	csk1->ka_timeout = DEF_KA_TIMEOUT;
	csk1->ka_interval = DEF_KA_INTERVAL;
	csk1->ka_max_probe_count = DEF_KA_MAX_PROBE_COUNT;
	csk1->tos = DEF_TOS;
	csk1->ttl = DEF_TTL;
	csk1->snd_seq_scale = DEF_SND_SEQ_SCALE;
	csk1->rcv_buf = DEF_RCV_BUF;
	csk1->snd_buf = DEF_SND_BUF;
	csk1->seed = DEF_SEED;

	*csk = csk1;

	return 0;
}

static int cnic_cm_destroy(struct cnic_sock *csk)
{
	struct cnic_local *cp = csk->dev->cnic_priv;

	if (!test_bit(SK_F_INUSE, &csk->flags))
		return -EINVAL;

	spin_lock_bh(&cp->cm_lock);
	if (csk->dst) {
		if (csk->dst->neighbour)
			neigh_release(csk->dst->neighbour);
		dst_release(csk->dst);
		csk->dst = NULL;
	}
	csk->flags = 0;
	spin_unlock_bh(&cp->cm_lock);
	return 0;
}

static inline struct net_device *get_real_netdev(struct net_device *netdev)
{
	return netdev->priv_flags & IFF_802_1Q_VLAN ?
		VLAN_DEV_INFO(netdev)->real_dev : netdev;
}

static struct cnic_dev *cnic_cm_select_dev(struct sockaddr_in *dst_addr,
					   int ulp_type)
{
	u32 dst_ip = dst_addr->sin_addr.s_addr;
	struct flowi fl;
	struct rtable *rt;
	struct net_device *netdev;
	struct cnic_dev *dev;
	int err, found;

	memset(&fl, 0, sizeof(fl));
	fl.nl_u.ip4_u.daddr = dst_ip;

	err = ip_route_output_key(&rt, &fl);
	if (err)
		return NULL;

	netdev = get_real_netdev(rt->idev->dev);

	found = 0;
	read_lock(&cnic_dev_lock);
	list_for_each_entry(dev, &cnic_dev_list, list) {
		if (netdev == dev->netdev) {
			found = 1;
			cnic_hold(dev);
			break;
		}
	}
	read_unlock(&cnic_dev_lock);

	ip_rt_put(rt);

	if (found) {
		struct cnic_local *cp = dev->cnic_priv;

		if (rcu_dereference(cp->ulp_ops[ulp_type])) {
			cnic_put(dev);
			return dev;
		}
		cnic_put(dev);
	}

	return NULL;
}

static int cnic_cm_connect(struct cnic_sock *csk, struct cnic_sockaddr *saddr)
{
	struct cnic_dev *dev = csk->dev;
	struct net_device *realdev;
	u32 dst_ip = saddr->remote_addr.sin_addr.s_addr;
	u32 src_ip = saddr->local_addr.sin_addr.s_addr;
	struct flowi fl;
	struct rtable *rt;
	struct neighbour *neigh;
	int err = 0;
#ifndef HAVE_NETEVENT
	int retry = 0;
#endif

	if (!test_bit(SK_F_INUSE, &csk->flags))
		return -EINVAL;

	memset(&fl, 0, sizeof(fl));
	fl.nl_u.ip4_u.daddr = dst_ip;
	fl.nl_u.ip4_u.saddr = src_ip;
	err = ip_route_output_key(&rt, &fl);
	if (err)
		return err;

	realdev = get_real_netdev(rt->idev->dev);
	if (realdev != dev->netdev)
		goto err;

	if (src_ip == 0)
		src_ip = inet_select_addr(rt->idev->dev, dst_ip, RT_SCOPE_LINK);

	csk->dst = &rt->u.dst;
	csk->src_ip = src_ip;
	csk->dst_ip = dst_ip;
	csk->src_port = saddr->local_addr.sin_port;
	csk->dst_port = saddr->remote_addr.sin_port;

	neigh = csk->dst->neighbour;
	if (!csk->dst->neighbour)
		goto err;

	neigh_hold(neigh);

	if (realdev != rt->idev->dev)
		csk->vlan_id = VLAN_DEV_INFO(rt->idev->dev)->vlan_id;
	else
		csk->vlan_id = 0;


	if (neigh->nud_state & NUD_CONNECTED) {
		set_bit(SK_F_PG_OFFLD_PENDING, &csk->flags);
		err = cnic_cm_offload_pg(csk);
		goto done;
	}
#ifdef HAVE_NETEVENT
	neigh_event_send(neigh, NULL);
#else
	while (!(neigh->nud_state & NUD_CONNECTED) && (retry < 3)) {
		arp_send(ARPOP_REQUEST, ETH_P_ARP, rt->rt_gateway,
			 rt->idev->dev, rt->rt_src, NULL,
			 rt->idev->dev->dev_addr, NULL);
		msleep(1000);
		retry++;
	}
	if (!(neigh->nud_state & NUD_CONNECTED))
		err = -ETIMEDOUT;
	else {
		set_bit(SK_F_PG_OFFLD_PENDING, &csk->flags);
		err = cnic_cm_offload_pg(csk);
	}
#endif

done:
	if (!err)
		return 0;

	neigh_release(neigh);

err:
	csk->dst = NULL;

	ip_rt_put(rt);
	return err;
}

static int cnic_cm_abort(struct cnic_sock *csk)
{
	if (!test_bit(SK_F_INUSE, &csk->flags))
		return -EINVAL;

	return (cnic_cm_abort_req(csk));
}

static int cnic_cm_close(struct cnic_sock *csk)
{
	if (!test_bit(SK_F_INUSE, &csk->flags))
		return -EINVAL;

	return (cnic_cm_close_req(csk));
}

static void cnic_cm_upcall(struct cnic_local *cp, struct cnic_sock *csk,
			   u8 opcode)
{
	struct cnic_ulp_ops *ulp_ops;
	int ulp_type = csk->ulp_type;

	rcu_read_lock();
	ulp_ops = rcu_dereference(cp->ulp_ops[ulp_type]);
	if (ulp_ops) {
		if (opcode == L4_KCQE_OPCODE_VALUE_CONNECT_COMPLETE)
			ulp_ops->cm_connect_complete(csk);
		else if (opcode == L4_KCQE_OPCODE_VALUE_CLOSE_COMP)
			ulp_ops->cm_close_complete(csk);
		else if (opcode == L4_KCQE_OPCODE_VALUE_RESET_RECEIVED)
			ulp_ops->cm_remote_abort(csk);
		else if (opcode == L4_KCQE_OPCODE_VALUE_RESET_COMP)
			ulp_ops->cm_abort_complete(csk);
		else if (opcode == L4_KCQE_OPCODE_VALUE_CLOSE_RECEIVED)
			ulp_ops->cm_remote_close(csk);
	}
	rcu_read_unlock();
}

static int cnic_cm_set_pg(struct cnic_sock *csk)
{
	struct neighbour *neigh = csk->dst->neighbour;
	int valid = neigh->nud_state & NUD_VALID;

	if (!valid) {
		if (!test_bit(SK_F_PG_OFFLD_PENDING, &csk->flags) &&
		    !test_bit(SK_F_PG_OFFLD_COMPLETE, &csk->flags))
			return -ETIMEDOUT;
	} else if (test_bit(SK_F_PG_OFFLD_COMPLETE, &csk->flags))
		cnic_cm_update_pg(csk);
	else if (!test_and_set_bit(SK_F_PG_OFFLD_PENDING, &csk->flags))
		cnic_cm_offload_pg(csk);
	return 0;
}

static void cnic_cm_process_neigh(struct cnic_dev *dev, struct neighbour *neigh)
{
	struct cnic_local *cp = dev->cnic_priv;
	int i;

	for (i = 0; i < MAX_CM_SK_TBL_SZ; i++) {
		struct cnic_sock *csk;
		int abort = 0;

		csk = &cp->csk_tbl[i];
		spin_lock_bh(&cp->cm_lock);
		if (test_bit(SK_F_INUSE, &csk->flags) && csk->dst &&
		    csk->dst->neighbour == neigh) {
			if (cnic_cm_set_pg(csk))
				abort = 1;
		}
		spin_unlock_bh(&cp->cm_lock);
		if (abort)
			cnic_cm_upcall(cp, csk,
				L4_KCQE_OPCODE_VALUE_CONNECT_COMPLETE);
	}
	neigh_release(neigh);
}

static void cnic_cm_process_redirect(struct cnic_dev *dev,
				     struct cnic_redirect_entry *redir)
{
	struct cnic_local *cp = dev->cnic_priv;
	int i;

	for (i = 0; i < MAX_CM_SK_TBL_SZ; i++) {
		struct cnic_sock *csk;
		int abort = 0;

		spin_lock_bh(&cp->cm_lock);
		csk = &cp->csk_tbl[i];
		if (test_bit(SK_F_INUSE, &csk->flags) &&
		    csk->dst == redir->old_dst) {
			csk->dst = redir->new_dst;
			dst_hold(csk->dst);
			neigh_hold(csk->dst->neighbour);
			if (redir->old_dst->neighbour);
				neigh_release(redir->old_dst->neighbour);
			dst_release(redir->old_dst);
			if (cnic_cm_set_pg(csk))
				abort = 1;
		}
		spin_unlock_bh(&cp->cm_lock);
		if (abort)
			cnic_cm_upcall(cp, csk,
				L4_KCQE_OPCODE_VALUE_CONNECT_COMPLETE);
	}

	dst_release(redir->new_dst);
}

static void cnic_close_conn(struct cnic_sock *csk)
{
	clear_bit(SK_F_OFFLD_COMPLETE, &csk->flags);
	clear_bit(SK_F_OFFLD_PENDING, &csk->flags);
	cnic_cm_upload_pg(csk);
	clear_bit(SK_F_PG_OFFLD_COMPLETE, &csk->flags);
}

static void cnic_cm_process_offld_pg(struct cnic_dev *dev, struct l4_kcq *kcqe)
{
	struct cnic_local *cp = dev->cnic_priv;
	u32 l5_cid = kcqe->pg_host_opaque;
	struct cnic_sock *csk = &cp->csk_tbl[l5_cid];

	if (!test_bit(SK_F_INUSE, &csk->flags))
		return;

	csk->pg_cid = kcqe->pg_cid;
	set_bit(SK_F_PG_OFFLD_COMPLETE, &csk->flags);
	clear_bit(SK_F_PG_OFFLD_PENDING, &csk->flags);
	if (cnic_ok_to_connect(csk)) {
		set_bit(SK_F_OFFLD_PENDING, &csk->flags);
		cnic_cm_conn_req(csk);
	}
}

static void cnic_cm_process_kcqe(struct cnic_dev *dev, struct kcqe *kcqe)
{
	struct cnic_local *cp = dev->cnic_priv;
	struct l4_kcq *l4kcqe = (struct l4_kcq *) kcqe;
	u8 opcode = l4kcqe->op_code;
	u32 l5_cid;
	struct cnic_sock *csk;

	if (opcode == L4_KCQE_OPCODE_VALUE_OFFLOAD_PG) {
		cnic_cm_process_offld_pg(dev, l4kcqe);
		return;
	}

	l5_cid = l4kcqe->conn_id;
	csk = &cp->csk_tbl[l5_cid];

	if (!test_bit(SK_F_INUSE, &csk->flags))
		return;

	if (opcode == L4_KCQE_OPCODE_VALUE_CONNECT_COMPLETE) {

		clear_bit(SK_F_OFFLD_PENDING, &csk->flags);
		if (l4kcqe->status == 0)
			set_bit(SK_F_OFFLD_COMPLETE, &csk->flags);
		cnic_cm_upcall(cp, csk, opcode);

	} else if (opcode == L4_KCQE_OPCODE_VALUE_CLOSE_COMP ||
		   opcode == L4_KCQE_OPCODE_VALUE_RESET_COMP ||
		   opcode == L4_KCQE_OPCODE_VALUE_RESET_RECEIVED) {

		cnic_close_conn(csk);
		cnic_cm_upcall(cp, csk, opcode);

	} else if (opcode == L4_KCQE_OPCODE_VALUE_CLOSE_RECEIVED) {
		cnic_cm_upcall(cp, csk, opcode);
	}
}

static void cnic_cm_indicate_kcqe(void *data, struct kcqe *kcqe[], u32 num_cqe)
{
	struct cnic_dev *dev = data;
	int i;
	struct cnic_local *cp = dev->cnic_priv;

	for (i = 0; i < num_cqe; i++)
		cnic_queue_work(cp, WORK_TYPE_KCQE, kcqe[i]);

	tasklet_schedule(&cp->cnic_task);
}

static void cnic_cm_indicate_event(void *data, unsigned long event)
{
}

static void cnic_cm_dummy(void *data)
{
}

static struct cnic_ulp_ops cm_ulp_ops = {
	.cnic_start		= cnic_cm_dummy,
	.cnic_stop		= cnic_cm_dummy,
	.indicate_kcqes		= cnic_cm_indicate_kcqe,
	.indicate_netevent	= cnic_cm_indicate_event,
	.indicate_inetevent	= cnic_cm_indicate_event,
};

static void cnic_task(unsigned long data)
{
	struct cnic_local *cp = (struct cnic_local *) data;
	struct cnic_dev *dev = cp->dev;
	u32 cons = cp->cnic_wr_cons;
	u32 prod = cp->cnic_wr_prod;

	while (cons != prod) {
		struct cnic_work_node *node;

		node = &cp->cnic_work_ring[cons];
		if (node->work_type == WORK_TYPE_KCQE)
			cnic_cm_process_kcqe(dev, &node->work_data.kcqe);
		else if (node->work_type == WORK_TYPE_NEIGH_UPDATE)
			cnic_cm_process_neigh(dev, node->work_data.neigh);
		else if (node->work_type == WORK_TYPE_REDIRECT)
			cnic_cm_process_redirect(dev,
				&node->work_data.cnic_redir);
		cons++;
		cons &= WORK_RING_SIZE_MASK;
	}
	cp->cnic_wr_cons = cons;
}

static void cnic_free_dev(struct cnic_dev *dev)
{
	int i = 0;

	while ((atomic_read(&dev->ref_count) != 0) && i < 10) {
		msleep(100);
		i++;
	}
	if (atomic_read(&dev->ref_count) != 0)
		printk(KERN_ERR PFX "%s: Failed waiting for ref count to go"
				    " to zero.\n", dev->netdev->name);

	printk(KERN_INFO PFX "Removed CNIC device: %s\n", dev->netdev->name);
	cnic_free_resc(dev);
	cnic_unregister_sysfs(dev);
	pci_dev_put(dev->pcidev);
	dev_put(dev->netdev);
	kfree(dev);
}

static void cnic_cm_free_mem(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;

	kfree(cp->csk_tbl);
	cp->csk_tbl = NULL;
}

static int cnic_cm_alloc_mem(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;

	cp->csk_tbl = kmalloc(sizeof(struct cnic_sock) * MAX_CM_SK_TBL_SZ,
			      GFP_KERNEL);
	if (!cp->csk_tbl)
		return -ENOMEM;
	memset(cp->csk_tbl, 0, sizeof(struct cnic_sock) * MAX_CM_SK_TBL_SZ);
	return 0;
}

#ifndef HAVE_NETEVENT
static void cnic_timer(unsigned long data)
{
	struct cnic_local *cp = (struct cnic_local *) data;
	struct cnic_dev *dev = cp->dev;
	int i, found = 0;
	struct neighbour *neigh = NULL;

	if (!test_bit(CNIC_F_CNIC_UP, &dev->flags))
		return;

	spin_lock_bh(&cp->cm_lock);
	for (i = 0; i < MAX_CM_SK_TBL_SZ; i++) {
		struct cnic_sock *csk;

		csk = &cp->csk_tbl[i];
		if (test_bit(SK_F_INUSE, &csk->flags) && csk->dst &&
		    test_bit(SK_F_PG_OFFLD_COMPLETE, &csk->flags)) {
			neigh = csk->dst->neighbour;
			if (memcmp(csk->old_ha, neigh->ha, 6)) {
				found = 1;
				break;
			}
		}
	}
	spin_unlock_bh(&cp->cm_lock);

	if (found) {
		neigh_hold(neigh);

		cnic_queue_work(cp, WORK_TYPE_NEIGH_UPDATE, &neigh);
		tasklet_schedule(&cp->cnic_task);
	}

	cp->cnic_timer.expires = jiffies + cp->cnic_timer_off;
	add_timer(&cp->cnic_timer);
}
#endif

static int cnic_cm_open(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;
	u32 seed;
	int err;

#if (LINUX_VERSION_CODE >= 0x020612)
	get_random_bytes(&seed, 4);
#else
	seed = 0x12345678;
#endif
	bnx2_ctx_wr(dev->bp, 45, 0, seed);

	err = cnic_cm_alloc_mem(dev);
	if (err) {
		cnic_cm_free_mem(dev);
		return err;
	}

	spin_lock_init(&cp->cm_lock);

	tasklet_init(&cp->cnic_task, &cnic_task, (unsigned long) cp);

#ifdef HAVE_NETEVENT
	cp->cm_nb.notifier_call = cnic_net_callback;
	register_netevent_notifier(&cp->cm_nb);
#else
	init_timer(&cp->cnic_timer);
	cp->cnic_timer_off = 2 * HZ;
	cp->cnic_timer.expires = jiffies + cp->cnic_timer_off;
	cp->cnic_timer.data = (unsigned long) cp;
	cp->cnic_timer.function = cnic_timer;
	add_timer(&cp->cnic_timer);
#endif

	dev->cm_create = cnic_cm_create;
	dev->cm_destroy = cnic_cm_destroy;
	dev->cm_connect = cnic_cm_connect;
	dev->cm_abort = cnic_cm_abort;
	dev->cm_close = cnic_cm_close;
	dev->cm_select_dev = cnic_cm_select_dev;

	cp->ulp_handle[CNIC_ULP_L4] = dev;
	rcu_assign_pointer(cp->ulp_ops[CNIC_ULP_L4], &cm_ulp_ops);
	return 0;
}

static void cnic_cm_cleanup(struct cnic_sock *csk)
{
	clear_bit(SK_F_INUSE, &csk->flags);
	if (csk->dst) {
		if (csk->dst->neighbour)
			neigh_release(csk->dst->neighbour);
		dst_release(csk->dst);
		csk->dst = NULL;
	}
}

static int cnic_cm_shutdown(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;
	int i;

#ifdef HAVE_NETEVENT
	unregister_netevent_notifier(&cp->cm_nb);
#else
	del_timer_sync(&cp->cnic_timer);
#endif

	tasklet_disable(&cp->cnic_task);

	if (!cp->csk_tbl)
		return 0;

	for (i = 0; i < MAX_CM_SK_TBL_SZ; i++) {
		struct cnic_sock *csk = &cp->csk_tbl[i];

		cnic_cm_cleanup(csk);
	}
	cnic_cm_free_mem(dev);

	return 0;
}

static void cnic_init_context(struct cnic_dev *dev, u32 cid)
{
	u32 cid_addr;
	int i;

	cid_addr = GET_CID_ADDR(cid);

	for (i = 0; i < CTX_SIZE; i += 4)
		bnx2_ctx_wr(dev->bp, cid_addr, i, 0);
}

static int cnic_setup_5709_context(struct cnic_dev *dev, int valid)
{
	struct cnic_local *cp = dev->cnic_priv;
	int ret = 0, i;
	u32 valid_bit = valid ? BNX2_CTX_HOST_PAGE_TBL_DATA0_VALID : 0;

	if (CHIP_NUM(cp) != CHIP_NUM_5709)
		return 0;

	for (i = 0; i < cp->ctx_pages; i++) {
		int j;
		u32 idx = cp->ctx_arr[i].cid / cp->cids_pp;
		u32 val;

		REG_WR(dev, BNX2_CTX_HOST_PAGE_TBL_DATA0,
		       (cp->ctx_arr[i].mapping & 0xffffffff) | valid_bit);
		REG_WR(dev, BNX2_CTX_HOST_PAGE_TBL_DATA1,
		       (u64) cp->ctx_arr[i].mapping >> 32);
		REG_WR(dev, BNX2_CTX_HOST_PAGE_TBL_CTRL, idx |
		       BNX2_CTX_HOST_PAGE_TBL_CTRL_WRITE_REQ);
		for (j = 0; j < 10; j++) {

			val = REG_RD(dev, BNX2_CTX_HOST_PAGE_TBL_CTRL);
			if (!(val & BNX2_CTX_HOST_PAGE_TBL_CTRL_WRITE_REQ))
				break;
			udelay(5);
		}
		if (val & BNX2_CTX_HOST_PAGE_TBL_CTRL_WRITE_REQ) {
			ret = -EBUSY;
			break;
		}
	}
	return ret;
}

#ifdef CONFIG_PCI_MSI
static int cnic_init_msix(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;
	struct bnx2 *bp = dev->bp;

	if (bp->flags & USING_MSIX_FLAG) {
		int err;

		REG_WR(dev, BNX2_HC_SB_CONFIG_8, BNX2_HC_SB_CONFIG_8_ONE_SHOT);

		REG_WR(dev, BNX2_HC_COMP_PROD_TRIP_8, (2 << 16) | 8);
		REG_WR(dev, BNX2_HC_COM_TICKS_8, (64 << 16) | 220);
		REG_WR(dev, BNX2_HC_CMD_TICKS_8, (64 << 16) | 220);

		cp->status_blk = (void *) ((unsigned long) bp->status_blk +
					    BNX2_SBLK_MSIX_ALIGN_SIZE *
					    BNX2_CNIC_VEC);
		cp->last_status_idx = cp->status_blk->status_idx;
		tasklet_init(&cp->cnic_msix_task, &cnic_service_kq_msix,
			     (unsigned long) dev);
		err = request_irq(bp->msix_entries[BNX2_CNIC_VEC].vector,
				  cnic_msix, 0,
				  bp->msix_tbl[BNX2_CNIC_VEC].name, dev);
		if (err) {
			tasklet_disable(&cp->cnic_msix_task);
			return err;
		}
	}
	return 0;
}

static void cnic_free_msix(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;
	struct bnx2 *bp = dev->bp;

	if (bp->flags & USING_MSIX_FLAG) {
		synchronize_irq(bp->msix_entries[BNX2_CNIC_VEC].vector);
		REG_WR(dev, BNX2_PCICFG_INT_ACK_CMD, (BNX2_CNIC_VEC << 24) |
		       BNX2_PCICFG_INT_ACK_CMD_MASK_INT);
		tasklet_disable(&cp->cnic_msix_task);
		free_irq(bp->msix_entries[BNX2_CNIC_VEC].vector, dev);
	}
}

static void cnic_enable_int(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;
	struct bnx2 *bp = dev->bp;

	if (!(bp->flags & USING_MSIX_FLAG))
		return;

	REG_WR(dev, BNX2_PCICFG_INT_ACK_CMD, (BNX2_CNIC_VEC << 24) |
	       BNX2_PCICFG_INT_ACK_CMD_INDEX_VALID | cp->last_status_idx);
}

static void cnic_disable_int_sync(struct cnic_dev *dev)
{
	struct bnx2 *bp = dev->bp;

	if (!(bp->flags & USING_MSIX_FLAG))
		return;

	REG_WR(dev, BNX2_PCICFG_INT_ACK_CMD, (BNX2_CNIC_VEC << 24) |
	       BNX2_PCICFG_INT_ACK_CMD_MASK_INT);
	REG_RD(dev, BNX2_PCICFG_INT_ACK_CMD);
	synchronize_irq(bp->msix_entries[BNX2_CNIC_VEC].vector);
}

#else

static int cnic_init_msix(struct cnic_dev *dev)
{
	return 0;
}

static void cnic_free_msix(struct cnic_dev *dev)
{
}

static void cnic_enable_int(struct cnic_dev *dev)
{
}

static void cnic_disable_int_sync(struct cnic_dev *dev)
{
}
#endif

static int cnic_start_hw(struct cnic_dev *dev)
{
	struct cnic_local *cp = dev->cnic_priv;
	struct bnx2 *bp = dev->bp;
	u32 val;
	int err;

	if (test_bit(CNIC_F_CNIC_UP, &dev->flags))
		return -EALREADY;

	val = REG_RD(dev, BNX2_MQ_CONFIG);
	val &= ~BNX2_MQ_CONFIG_KNL_BYP_BLK_SIZE;
	if (BCM_PAGE_BITS > 12)
		val |= (12 - 8)  << 4;
	else
		val |= (BCM_PAGE_BITS - 8)  << 4;

	REG_WR(dev, BNX2_MQ_CONFIG, val);

	REG_WR(dev, BNX2_HC_COMP_PROD_TRIP, (2 << 16) | 8);
	REG_WR(dev, BNX2_HC_COM_TICKS, (64 << 16) | 220);
	REG_WR(dev, BNX2_HC_CMD_TICKS, (64 << 16) | 220);

	err = cnic_setup_5709_context(dev, 1);
	if (err)
		return err;

	cnic_init_context(dev, KWQ_CID);
	cnic_init_context(dev, KCQ_CID);

	cp->kwq_cid_addr = GET_CID_ADDR(KWQ_CID);
	cp->kwq_io_addr = MB_GET_CID_ADDR(KWQ_CID) + L5_KRNLQ_HOST_QIDX;

	cnic_setup_page_tbl(cp->kwq_pgtbl, KWQ_PAGE_CNT, cp->kwq_mapping);
	
	cp->kwq_prod_idx = 0;
	cp->kwq_con_idx = 0;
	
	/* Initialize the kernel work queue context. */
	val = KRNLQ_TYPE_TYPE_KRNLQ | KRNLQ_SIZE_TYPE_SIZE |
	      (BCM_PAGE_BITS - 8) | KRNLQ_FLAGS_QE_SELF_SEQ;
	bnx2_ctx_wr(bp, cp->kwq_cid_addr, L5_KRNLQ_TYPE, val);
	
	val = (PAGE_SIZE / sizeof(struct kwqe) - 1) << 16;
	bnx2_ctx_wr(bp, cp->kwq_cid_addr, L5_KRNLQ_QE_SELF_SEQ_MAX, val);
	
	val = ((PAGE_SIZE / sizeof(struct kwqe)) << 16) | KWQ_PAGE_CNT;
	bnx2_ctx_wr(bp, cp->kwq_cid_addr, L5_KRNLQ_PGTBL_NPAGES, val);
	
	val = (u32) ((u64) cp->kwq_pgtbl_mapping >> 32);
	bnx2_ctx_wr(bp, cp->kwq_cid_addr, L5_KRNLQ_PGTBL_HADDR_HI, val);
	
	val = (u32) cp->kwq_pgtbl_mapping;
	bnx2_ctx_wr(bp, cp->kwq_cid_addr, L5_KRNLQ_PGTBL_HADDR_LO, val);

	cp->kcq_cid_addr = GET_CID_ADDR(KCQ_CID);
	cp->kcq_io_addr = MB_GET_CID_ADDR(KCQ_CID) + L5_KRNLQ_HOST_QIDX;
	
	cnic_setup_page_tbl(cp->kcq_pgtbl, KCQ_PAGE_CNT, cp->kcq_mapping);
	
	cp->kcq_prod_idx = 0;
	
	/* Initialize the kernel complete queue context. */
	val = KRNLQ_TYPE_TYPE_KRNLQ | KRNLQ_SIZE_TYPE_SIZE |
	      (BCM_PAGE_BITS - 8) | KRNLQ_FLAGS_QE_SELF_SEQ;
	bnx2_ctx_wr(bp, cp->kcq_cid_addr, L5_KRNLQ_TYPE, val);
	
	val = (BCM_PAGE_SIZE / sizeof(struct kcqe) - 1) << 16;
	bnx2_ctx_wr(bp, cp->kcq_cid_addr, L5_KRNLQ_QE_SELF_SEQ_MAX, val);
	
	val = ((BCM_PAGE_SIZE / sizeof(struct kcqe)) << 16) | KCQ_PAGE_CNT;
	bnx2_ctx_wr(bp, cp->kcq_cid_addr, L5_KRNLQ_PGTBL_NPAGES, val);
	
	val = (u32) ((u64) cp->kcq_pgtbl_mapping >> 32);
	bnx2_ctx_wr(bp, cp->kcq_cid_addr, L5_KRNLQ_PGTBL_HADDR_HI, val);
	
	val = (u32) cp->kcq_pgtbl_mapping;
	bnx2_ctx_wr(bp, cp->kcq_cid_addr, L5_KRNLQ_PGTBL_HADDR_LO, val);

	if (bp->flags & USING_MSIX_FLAG) {
		val = KRNLQ_STATUSB_NUM_CNIC_MSIX << KRNLQ_STATUSB_NUM_SHIFT;
		bnx2_ctx_wr(bp, cp->kwq_cid_addr, L5_KRNLQ_HOST_QIDX, val);
		bnx2_ctx_wr(bp, cp->kcq_cid_addr, L5_KRNLQ_HOST_QIDX, val);
	}

	/* Enable Commnad Scheduler notification when we write to the
	 * host producer index of the kernel contexts. */
	REG_WR(dev, BNX2_MQ_KNL_CMD_MASK1, 2);

	/* Enable Command Scheduler notification when we write to either
	 * the Send Queue or Receive Queue producer indexes of the kernel
	 * bypass contexts. */
	REG_WR(dev, BNX2_MQ_KNL_BYP_CMD_MASK1, 7);
	REG_WR(dev, BNX2_MQ_KNL_BYP_WRITE_MASK1, 7);

	/* Notify COM when the driver post an application buffer. */
	REG_WR(dev, BNX2_MQ_KNL_RX_V2P_MASK2, 0x2000);

	err = cnic_init_msix(dev);
	if (err) {
		printk(KERN_ERR PFX "%s: cnic_init_msix failed\n",
		       dev->netdev->name);
		return err;
	}

	/* Set the CP and COM doorbells.  These two processors polls the
	 * doorbell for a non zero value before running.  This must be done
	 * after setting up the kernel queue contexts. */
	bnx2_reg_wr_ind(bp, BNX2_CP_SCRATCH + 0x20, 1);
	bnx2_reg_wr_ind(bp, BNX2_COM_SCRATCH + 0x20, 1);

	err = bnx2_register_cnic(dev->netdev, &my_cnic_ops, dev);
	if (err) {
		printk(KERN_ERR PFX "%s: bnx2_register_cnic failed\n",
		       dev->netdev->name);
		bnx2_reg_wr_ind(dev->bp, BNX2_CP_SCRATCH + 0x20, 0);
		bnx2_reg_wr_ind(dev->bp, BNX2_COM_SCRATCH + 0x20, 0);
		return -EBUSY;
	}

	set_bit(CNIC_F_CNIC_UP, &dev->flags);
	cnic_cm_open(dev);

	cnic_enable_int(dev);

	return 0;
}

static void cnic_stop_hw(struct cnic_dev *dev)
{
	if (test_bit(CNIC_F_CNIC_UP, &dev->flags)) {
		struct cnic_local *cp = dev->cnic_priv;

		clear_bit(CNIC_F_CNIC_UP, &dev->flags);
		rcu_assign_pointer(cp->ulp_ops[CNIC_ULP_L4], NULL);
		synchronize_rcu();
		cnic_cm_shutdown(dev);
		bnx2_unregister_cnic(dev->netdev);

		cnic_disable_int_sync(dev);

		bnx2_reg_wr_ind(dev->bp, BNX2_CP_SCRATCH + 0x20, 0);
		bnx2_reg_wr_ind(dev->bp, BNX2_COM_SCRATCH + 0x20, 0);

		cnic_init_context(dev, KWQ_CID);
		cnic_init_context(dev, KCQ_CID);

		cnic_setup_5709_context(dev, 0);
		cnic_free_msix(dev);
	}
}

static struct cnic_dev *init_cnic(struct net_device *dev)
{
	struct cnic_dev *cdev;
	struct cnic_local *cp;
	struct bnx2 *bnx2_bp = netdev_priv(dev);
	struct pci_dev *pdev = bnx2_bp->pdev;
	int alloc_size;

	if (!pdev)
		return NULL;

	pci_dev_get(pdev);
	if (pdev->device == PCI_DEVICE_ID_NX2_5709 ||
	    pdev->device == PCI_DEVICE_ID_NX2_5709S) {
		pci_dev_put(pdev);
		return NULL;
	}

	alloc_size = sizeof(struct cnic_dev) + sizeof(struct cnic_local);

	cdev = kmalloc(alloc_size , GFP_KERNEL);
	if (cdev == NULL) {
		printk(KERN_ERR PFX "%s: allocate dev struct failure\n",
		       dev->name);
		pci_dev_put(pdev);
		return cdev;
	}
	memset(cdev, 0, alloc_size);

	cdev->netdev = dev;
	cdev->bp = bnx2_bp;
	cdev->pcidev = pdev;
	cdev->regview = bnx2_bp->regview;
	cdev->cnic_priv = (char *)cdev + sizeof(struct cnic_dev);
	cdev->submit_kwqes = cnic_submit_kernel_wqes;
	cdev->register_device = cnic_register_device;
	cdev->unregister_device = cnic_unregister_device;
	cp = cdev->cnic_priv;
	cp->dev = cdev;
	cp->chip_id = bnx2_bp->chip_id;

	if (cnic_alloc_resc(cdev)) {
		printk(KERN_ERR PFX "%s: allocate resource failure\n",
		       dev->name);
		kfree(cdev);
		pci_dev_put(pdev);
		return NULL;
	}
	spin_lock_init(&cp->cnic_ulp_lock);
	dev_hold(cdev->netdev);
	printk(KERN_INFO PFX "Added CNIC device: %s\n", dev->name);

	return cdev;
}

static struct cnic_dev *is_cnic_dev(struct net_device *dev)
{
	struct ethtool_drvinfo drvinfo;
	struct cnic_dev *cdev = NULL;

	if (dev->ethtool_ops && dev->ethtool_ops->get_drvinfo) {
		memset(&drvinfo, 0, sizeof(drvinfo));
		dev->ethtool_ops->get_drvinfo(dev, &drvinfo);
		
		if (!strcmp(drvinfo.driver, "bnx2")) {
			cdev = init_cnic(dev);
			if (cdev) {
				cnic_register_sysfs(cdev);
				write_lock(&cnic_dev_lock);
				list_add(&cdev->list, &cnic_dev_list);
				write_unlock(&cnic_dev_lock);
			}
		}
	}
	return cdev;
}

/**
 * IP event handler
 */
static int cnic_ip_event(struct notifier_block *this, unsigned long event,
						 void *ptr)
{
	struct in_ifaddr *ifa = (struct in_ifaddr *) ptr;
	struct net_device *netdev = (struct net_device *) ifa->ifa_dev->dev;
	struct cnic_dev *dev;
	int if_type;
	u32 my_dev = 0;

	read_lock(&cnic_dev_lock);
	list_for_each_entry(dev, &cnic_dev_list, list) {
		if (netdev == dev->netdev) {
			my_dev = 1;
			cnic_hold(dev);
			break;
		}
	}
	read_unlock(&cnic_dev_lock);

	if (my_dev) {
		struct cnic_local *cp = dev->cnic_priv;

		rcu_read_lock();
		for (if_type = 0; if_type < MAX_CNIC_ULP_TYPE; if_type++) {
			struct cnic_ulp_ops *ulp_ops;

			ulp_ops = rcu_dereference(cp->ulp_ops[if_type]);
			if (ulp_ops) {
				void *ctx = cp->ulp_handle[if_type];

				ulp_ops->indicate_inetevent(ctx, event);
			}
		}
		rcu_read_unlock();

		cnic_put(dev);
	}

	return NOTIFY_DONE;
}

/**
 * netdev event handler
 */
static int cnic_netdev_event(struct notifier_block *this, unsigned long event,
							 void *ptr)
{
	struct net_device *netdev = ptr;
	struct cnic_dev *dev;
	int if_type;
	u32 my_dev = 0;

	read_lock(&cnic_dev_lock);
	list_for_each_entry(dev, &cnic_dev_list, list) {
		if (netdev == dev->netdev) {
			my_dev = 1;
			cnic_hold(dev);
			break;
		}
	}
	read_unlock(&cnic_dev_lock);

	if (!my_dev && event == NETDEV_REGISTER) {
		/* Check for the hot-plug device */
		dev = is_cnic_dev(netdev);
		if (dev) {
			my_dev = 1;
			cnic_hold(dev);
		}
	}
	if (my_dev) {
		struct cnic_local *cp = dev->cnic_priv;

		if (event == NETDEV_REGISTER)
			cnic_ulp_init(dev);
		else if (event == NETDEV_UNREGISTER)
			cnic_ulp_exit(dev);
		else if (event == NETDEV_UP) {
			mutex_lock(&cnic_lock);
			set_bit(CNIC_F_IF_UP, &dev->flags);
			if (dev->use_count) {
				if (!cnic_start_hw(dev))
					cnic_ulp_start(dev);
			}
			mutex_unlock(&cnic_lock);
		}

		rcu_read_lock();
		for (if_type = 0; if_type < MAX_CNIC_ULP_TYPE; if_type++) {
			struct cnic_ulp_ops *ulp_ops;
			void *ctx;

			ulp_ops = rcu_dereference(cp->ulp_ops[if_type]);
			if (!ulp_ops)
				continue;

			ctx = cp->ulp_handle[if_type];

			ulp_ops->indicate_netevent(ctx, event);
		}
		rcu_read_unlock();

		if (event == NETDEV_GOING_DOWN) {
			mutex_lock(&cnic_lock);
			clear_bit(CNIC_F_IF_UP, &dev->flags);
			set_bit(CNIC_F_IF_GOING_DOWN, &dev->flags);
			cnic_ulp_stop(dev);
			cnic_stop_hw(dev);
			mutex_unlock(&cnic_lock);
		} else if (event == NETDEV_DOWN) {
			mutex_lock(&cnic_lock);
			clear_bit(CNIC_F_IF_GOING_DOWN, &dev->flags);
			mutex_unlock(&cnic_lock);
		} else if (event == NETDEV_UNREGISTER) {
			write_lock(&cnic_dev_lock);
			list_del_init(&dev->list);
			write_unlock(&cnic_dev_lock);

			cnic_put(dev);
			cnic_free_dev(dev);
			goto done;
		}
		cnic_put(dev);
	}
done:
	return NOTIFY_DONE;
}

static struct notifier_block cnic_ip_notifier = {
	cnic_ip_event,
	0
};

static struct notifier_block cnic_netdev_notifier = {
	cnic_netdev_event,
	0
};

static void cnic_release(void)
{
	struct cnic_dev *dev;

	while (!list_empty(&cnic_dev_list)) {
		dev = list_entry(cnic_dev_list.next, struct cnic_dev, list);
		if (test_bit(CNIC_F_CNIC_UP, &dev->flags))
			cnic_stop_hw(dev);

		list_del_init(&dev->list);
		cnic_free_dev(dev);
	}
}

static int __init cnic_init(void)
{
	int rc = 0;
	struct net_device *dev;

	printk(KERN_INFO "%s", version);

	cnic_sysfs_setup();

	read_lock(&dev_base_lock);
	/* Find Teton devices */
#if (LINUX_VERSION_CODE >= 0x020618)
	for_each_netdev(&init_net, dev)
#elif (LINUX_VERSION_CODE >= 0x20616)
	for_each_netdev(dev)
#else
	for (dev = dev_base; dev; dev = dev->next)
#endif
		is_cnic_dev(dev);

	read_unlock(&dev_base_lock);

	rc = register_inetaddr_notifier(&cnic_ip_notifier);
	if (rc)
		cnic_release();
	rc = register_netdevice_notifier(&cnic_netdev_notifier);
	if (rc) {
		unregister_inetaddr_notifier(&cnic_ip_notifier);
		cnic_release();
	}
	return rc;
}

static void __exit cnic_exit(void)
{
	unregister_inetaddr_notifier(&cnic_ip_notifier);
	unregister_netdevice_notifier(&cnic_netdev_notifier);
	cnic_release();
	cnic_sysfs_cleanup();

	return;
}

module_init(cnic_init);
module_exit(cnic_exit);
